#!/bin/sh

PATH="/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/sbin:/usr/local/bin"

imgfile="ramdisk.img"
imgsize="4096"
fsnodes="256"
srcpath="src"
srcfile="src.tgz"
srcver="${srcpath}/etc/version"
srcdbgf="${srcpath}/etc/debug"
instmod="install -p -o root -g root -m 444"
modpath="modules"
moddest="${srcpath}/lib/modules"
instexe="install -p -o root -g root -m 555"
binpath="bin"
bindest="${srcpath}/bin"
outfile="ramdisk"

state="clean"

die() {
  echo "$0: ERROR: $1!"
  case $state in
  losetup)
    losetup -d ${loopdev}
    ;;
  mounted|unpacked)
    umount ${loopdev} && losetup -d ${loopdev}
    ;;
  esac
  exit 1
}

find_loopdev() {
  for i in /dev/loop*; do
    if test -b $i -a -r $i -a -w $i && ! losetup $i > /dev/null 2>&1; then
      loopdev=$i
      return
    fi
  done
  die "No free loop devices"
}

if test "x$1" = "x" -o "x$1" = "xrelease"; then
  dbgmode="n"
elif test "x$1" = "xdebug"; then
  dbgmode="y"
else
  echo "Usage: $0 [debug|release]"
  exit 1
fi

echo "**** Checking for root"
if ! test "`id -u`" == "0"; then
  die "You must be root to run this script"
fi

echo "**** Finding free loop device"
find_loopdev

echo "**** Creating ramdisk image"
if ! dd if=/dev/zero of=${imgfile} bs=1k count=${imgsize} > /dev/null 2>&1; then
  die "Unable to create ramdisk image"
fi

echo "**** Attaching loopback device"
if ! losetup ${loopdev} ${imgfile}; then
  die "Unable to attach loopback device"
fi

state="losetup"

echo "**** Creating filesystem"
if ! mke2fs -N ${fsnodes} -m 0 -L barcelona ${loopdev} > /dev/null 2>&1 \
|| ! tune2fs -c 0 -i 0 -U clear ${loopdev} > /dev/null 2>&1; then
  die "Unable to create filesystem"
fi

echo "**** Mounting filesystem"
if ! mkdir -p ${srcpath} || ! mount ${loopdev} ${srcpath}; then
  die "Unable to mount filesystem"
fi

state="mounted"

echo "**** Unpacking source file"
if ! tar xzf ${srcfile}; then
  die "Unable to unpack source file"
fi

state="unpacked"

echo "**** Adding timestamp"
if ! date "+%Y-%m-%d %H:%M:%S" > ${srcver}; then
  die "Unable to add timestamp"
fi

if test "${dbgmode}" = "y"; then
  echo "**** Adding debug flag"
  if ! touch ${srcdbgf}; then
    die "Unable to add debug flag"
  fi
fi

echo "**** Copying modules"
if ! ${instmod} ${modpath}/*.o ${moddest}; then
  die "Unable to copy modules"
fi

echo "**** Copying executables"
if ! ${instexe} ${binpath}/* ${bindest}; then
  die "Unable to copy executables"
fi

echo "**** Free ramdisk space left:"
if ! df ${loopdev} || ! df -i ${loopdev}; then
  die "Unable to show free space"
fi

echo "**** Unmounting filesystem"
if ! umount ${loopdev}; then
  die "Unable to unmount filesystem"
fi

state="losetup"

echo "**** Detaching loopback"
if ! losetup -d ${loopdev}; then
  die "Unable to detach loopback"
fi

state="clean"

echo "**** Compressing ramdisk image"
if ! gzip -9cn ${imgfile} > ${outfile}; then
  die "Unable to compress ramdisk image"
fi

echo "**** Cleaning up"
if ! rm -rf ${imgfile} ${srcpath}; then
  die "Unable to clean up"
fi

echo "**** Done"
